package ResponseCache.ResponseCache;

import Bytes.Bytes.ByteArray;
import Bytes.Bytes.ByteArrayReference;
import DataStructures.Array.Structures.Data;
import ProcessingUnit.ProcessingUnit.ProcessingUnitStructure;
import ProcessingUnit.ProcessingUnitServer.ProcessingUnitServerStructure;
import references.references.NumberReference;

import static Bytes.Bytes.Bytes.*;
import static DataStructures.Array.Structures.Structures.*;
import static JSON.Writer.Writer.WriteJSON;
import static ProcessingUnit.ProcessingUnit.ProcessingUnit.*;
import static ProcessingUnit.ProcessingUnitServer.ProcessingUnitServer.*;
import static Unicode.Unicode.Unicode.UTF16ToUTF8;


public class ResponseCache {
    public static boolean ResponseCacheInitialize(ProcessingUnitServerStructure client, ProcessingUnitServerStructure adminPort, ProcessingUnitStructure server, double elements, double maxElementSize, ResponseCacheState state) {
        double i;

        state.elements = elements;
        state.maxElementSize = maxElementSize;

        state.cachedElements = new CachedElement[(int)elements];
        for(i = 0d; i < elements; i = i + 1d){
            state.cachedElements[(int)i] = new CachedElement();
            state.cachedElements[(int)i].cached = false;
        }

        state.idles = 0d;
        state.requests = 0d;
        state.cached = 0d;
        state.duplicateHash = 0d;
        state.responseTooLargeToCache = 0d;

        return true;
    }

    public static void ResponseCacheIteration(ProcessingUnitServerStructure client, ProcessingUnitServerStructure adminPort, ProcessingUnitStructure server, ResponseCacheState state, NumberReference idleRef){
        ByteArrayReference request, response;
        double hash;
        CachedElement cachedElement;
        boolean cached, clientReq, adminPortReq;
        double [] data;
        Data adminPortOutput;
        char [] adminPortOutputJSON;
        ByteArray adminPortOutputBytes;
        ByteArrayReference adminPortRequestBytes;

        request = new ByteArrayReference();
        response = new ByteArrayReference();

        if(ServerCheck(client)) {
            clientReq = true;

            state.requests = state.requests + 1d;

            ServerReceive(client, request);

            data = ByteArrayToNumberArray(request.bytes);
            hash = ModulusHash(data, state.elements);
            cachedElement = state.cachedElements[(int) hash];

            cached = false;
            if (cachedElement.cached) {
                if (ByteArraysEqual(request.bytes, cachedElement.request)) {
                    cached = true;
                    state.cached = state.cached + 1d;
                }else{
                    state.duplicateHash = state.duplicateHash + 1d;
                }
            }

            if (!cached) {
                Call(server, request.bytes, response);
                if (ByteArrayLength(response.bytes) <= state.maxElementSize) {
                    if (cachedElement.cached) {
                        delete(cachedElement.request);
                        delete(cachedElement.response);
                    } else {
                        cachedElement.cached = true;
                    }
                    cachedElement.request = CopyByteArray(request.bytes);
                    cachedElement.response = CopyByteArray(response.bytes);
                }else{
                    state.responseTooLargeToCache = state.responseTooLargeToCache + 1d;
                }
            } else {
                response.bytes = CopyByteArray(cachedElement.response);
            }

            ServerSend(client, response.bytes);
        }else{
            clientReq = false;
        }

        if(ServerCheck(adminPort)) {
            adminPortReq = true;

            adminPortRequestBytes = new ByteArrayReference();
            ServerReceive(adminPort, adminPortRequestBytes);

            adminPortOutput = CreateStructData();
            AddNumberToStruct(adminPortOutput.structure, "idles".toCharArray(), state.idles);
            AddNumberToStruct(adminPortOutput.structure, "Client Reqs".toCharArray(), state.requests);
            AddNumberToStruct(adminPortOutput.structure, "Cached Reqs".toCharArray(), state.cached);
            AddNumberToStruct(adminPortOutput.structure, "Duplicate Hash Req".toCharArray(), state.duplicateHash);
            AddNumberToStruct(adminPortOutput.structure, "Over Cache Max Req".toCharArray(), state.responseTooLargeToCache);


            adminPortOutputJSON = WriteJSON(adminPortOutput);

            adminPortOutputBytes = UTF16ToUTF8(adminPortOutputJSON);

            ServerSend(adminPort, adminPortOutputBytes);

            FreeData(adminPortOutput);
            delete(adminPortOutputJSON);
            FreeByteArray(adminPortOutputBytes);
        }else{
            adminPortReq = false;
        }

        if(!clientReq && !adminPortReq){
            idleRef.numberValue = 1d;
            state.idles = state.idles + 1d;
        }else{
            idleRef.numberValue = 0d;
        }

        delete(request);
        delete(response);
    }

    public static double ModulusHash(double[] array, double m) {
        double hash, i;

        hash = 0d;

        for(i = 0d; i < array.length; i = i + 1d){
            hash = (hash + array[(int)i]) % m;
        }

        return hash;
    }

    public static void delete(Object o) {
        // Java has garbage collection.
    }
}







